﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace cleanac3
{
    class Ac3ExtractorFixer
    {
        public string audio_extension = "ac3";
        public double fps30 = 1 / (30.0 / 1.001);
        public double ac3 = .032;


        public static int MyBuffer = 1000;
        public ts_info[] data = new ts_info[MyBuffer];	// TS packets Buffer
        public ts_info[] ac3_unit = new ts_info[20];     // Hold an ac3 unit for padding
        public long pack_nr;	// Curent packet

        string f_name; // Input stream file name
        string outdir; // where to put the file
        public double fps30_total;
        public double ac3_total;
        public bool ac3_off = false;
        public double ac3_adjusted = 0.00;
        public long PTS33;
        public double PTS;

        public FileStream file; // Input stream
        public long max;	// File size in bytes
        public long poz;	// Curent index in bytes readed
        public byte[] rd = new Byte[188]; // Raw data
        public long nr_packs;


        public Ac3ExtractorFixer(string inpath, string output_dir)
        {
            this.f_name = inpath;
            this.outdir = output_dir;
            init();
        }

        public void init()
        {
            file = new FileStream(f_name, FileMode.Open, FileAccess.Read );	// Input File Stream
            int i = 0;
            for (i = 0; i < MyBuffer; i++) data[i] = new ts_info();
            max = file.Length;
            poz = 0;
            pack_nr = 0;
            nr_packs = max / 188;
            i = fill_buffer();
        }

        public int fill_buffer()
        {
            int i = 0, k;
            while ((i < MyBuffer) && (poz < max))
            {
                k = file.Read(rd, 0, 188);
                poz += 188;
                data[i].copy_ts(rd, 188);
                i++;
            }
            return i;
        }

        public void display_pack()
        {
            //TODO, if wanted
        }

        public void log(string txt)
        {
            Console.WriteLine(txt);
        }

        public void process(Boolean doh264, Boolean doac3)
        {
            bool need_ac3_unit = true;
            int DisconFound = 0;
            int[] vpos = new int[50];
            int[] apos = new int[50];
            int vndx = 0;
            int vfrm = 0;
            int aunit = 0;
            int lastcount = 0;
            long save;
            int cnt = 0;
            int x, i, p;
            bool stop;

            // this will run much faster if the input and output are on different physical drives
            // using a separate output directory variable allows for that.
            // also, having just one output file name makes batching rather impractical - base it on the input instead
            string AC3Cut = Path.Combine(Path.GetDirectoryName(this.outdir), Path.GetFileNameWithoutExtension(f_name) + "." + audio_extension);
            string H264 = Path.Combine(Path.GetDirectoryName(this.outdir), Path.GetFileNameWithoutExtension(f_name) + ".264");
            string MyTSInfo = Path.Combine(Path.GetDirectoryName(this.outdir), "tsinfo.txt");
            string MyStamps = Path.Combine(Path.GetDirectoryName(this.outdir), "mystamps.txt");
            StreamWriter f = new StreamWriter(MyTSInfo);
            StreamWriter stamps = new StreamWriter(MyStamps);

            double AC3_PTS = 0.0;
            double Vid_PTS = 0.0;
            int FrameOpen = 0;
            bit bt = new bit();
            //            FileStream cut_ts = new FileStream(MyCut, FileMode.Create);
            FileStream new_ac3 = (doac3)? new FileStream(AC3Cut, FileMode.Create):null;
            FileStream new_264 = (doh264)? new FileStream(H264, FileMode.Create):null;
            try
            {
                f.WriteLine("Transport Stream : " + file.Name);
                f.WriteLine("File size is : " + max.ToString());
                f.WriteLine("Estimated number of packets : " + nr_packs.ToString());
                f.WriteLine("************************************************************************");
                f.WriteLine("");
                save = pack_nr;
                pack_nr = 1;
                poz = 0;
                file.Position = 0;
                x = fill_buffer();
                stop = false;
                while (!stop)
                {
                    for (i = 0; i < x; i++)
                    {
                        //				f.WriteLine("Packet no. : "+pack_nr.ToString());
                        data[i].analyze_pack();
                        if (data[i].pack.data[0] != 71)
                        {
                            f.WriteLine("Error ... Sync byte not found ... Illegal stream format");
                            stop = true;
                            pack_nr--;
                            display_pack();
                            pack_nr++;
                            log("File tsinfo.txt created in program directory");
                            log(pack_nr.ToString() + "  packets written ok");
                            f.Close();
                            FileStream f1 = new FileStream(MyTSInfo, FileMode.Open);
                            log("File size is  " + f1.Length.ToString() + " bytes");
                            f1.Close();
                            return;
                        }
                        else
                        {
                            int k = data[i].pid;
                            if (data[i].pid == 0x001F)
                            {
                                if (lastcount != (data[i].count))
                                {
                                    vpos[vndx] = vfrm;
                                    vndx++;
                                    log("Frame discon: " + vfrm.ToString() + " (" + pack_nr.ToString() + ")");
                                    ac3_total += .100; // Discontinuities create a -100ms delay, so compensate for it here
                                    //                            ac3_total += ac3;
                                    f.WriteLine("************** Creating 100ms Correction **********");
                                    DisconFound = 1;
                                }
                                if (data[i].count == 15)
                                    lastcount = 0;
                                else
                                    lastcount = data[i].count + 1;
                            }
                            if ((k == 0x1011 || k == 0x1100) && data[i].pes_st)
                            {
                                if (data[i].pack.data[4] == 0x0 && data[i].pack.data[5] == 0x0 && data[i].pack.data[6] == 0x1)
                                {
                                    int afield = bt.ret_bit_value(data[i].pack.data[11], 0, 1);
                                    if (afield == 3 || afield == 1 || afield == 2)
                                    {
                                        int tmpval = (int)data[i].pack.data[13] >> 1 & 0x7;
                                        PTS33 = tmpval * (long)Math.Pow(2, 30);
                                        PTS33 += bt.ret_bit_value(data[i].pack.data[14], 0, 7) * (long)Math.Pow(2, 22);
                                        PTS33 += bt.ret_bit_value(data[i].pack.data[15], 0, 6) * (long)Math.Pow(2, 15);
                                        PTS33 += bt.ret_bit_value(data[i].pack.data[16], 0, 7) * (long)Math.Pow(2, 7);
                                        PTS33 += bt.ret_bit_value(data[i].pack.data[17], 0, 6);
                                        //HextextBox.Text += "\r\n -------------\r\n" + "PTS33: " + PTS33.ToString();
                                        PTS = Convert.ToDouble(PTS33) / 90000;
                                        //HextextBox.Text += "  PTS: " + PTS.ToString("F4");
                                    }
                                }
                            }



                            if (data[i].pid == 0x1011 && data[i].pes_st)
                            {
                                if (DisconFound == 1)
                                {
                                    ac3_total -= .100; // Undo the correction, we found video before audio, 
                                    f.WriteLine("**************Undoing 100ms Correction**********");
                                    DisconFound = 0;
                                }
                                if (vfrm == 0)
                                {
                                    Vid_PTS = PTS;
                                    log(((AC3_PTS - Vid_PTS) * 1000.0).ToString("F0") + " ms Delay");
                                }
                                vfrm++;
                                fps30_total += fps30;
                                string framestats = vfrm.ToString("D6") + " -- " + (fps30_total - ac3_total).ToString("F3") + " -- " + (fps30_total - ac3_total - ac3_adjusted).ToString("F3") + " -- " + ((vfrm * fps30).ToString("F1"));
                                //                       if(vfrm < 10000)
                                f.WriteLine(framestats);
                            }
                            if (data[i].pid == 0x1011 && (data[i].pes_st || data[i].AF))
                            {
                                if (data[i].AF)
                                    FrameOpen = 0;
                                if (FrameOpen == 1 && data[i].pes_st)
                                {
                                    //                                                vfrm++;
                                    //                                                fps30_total += fps30;
                                    log("Open Frame at " + vfrm.ToString());
                                }
                                if (data[i].pes_st)
                                    FrameOpen = 1;
                            }
                            if (data[i].pid == 0x1100 && need_ac3_unit)
                            {

                                AC3_PTS = PTS;
                                ac3_unit[cnt] = new ts_info();
                                ac3_unit[cnt].copy_ts(data[i].pack.data, 188);
                                ac3_unit[cnt].Ad_len = data[i].Ad_len;
                                ac3_unit[cnt].Pay_len = data[i].Pay_len;
                                if (data[i].pes_st)
                                {
                                    //                            ac3_unit[cnt].Ad_len = 20;
                                    ac3_unit[cnt].Pay_len -= 17;
                                }
                                cnt++;
                                if (data[i].AF == true)
                                    need_ac3_unit = false;
                            }
                            if (data[i].pid == 0x1100 && data[i].pes_st)
                            {
                                DisconFound = 0;
                                aunit++;
                                ac3_total += ac3;
                                if ((fps30_total - (ac3_total + ac3_adjusted)) > .024)
                                {
                                    ac3_adjusted += ac3;
                                    f.WriteLine("************* 32ms Added **********");

                                    for (int ac3_count = 0; ac3_count < cnt; ac3_count++)
                                    {
                                        //                                cut_ts.Write(ac3_unit[ac3_count].pack.data, 0, 188);
                                        if (new_ac3 !=null)
                                            new_ac3.Write(ac3_unit[ac3_count].pack.data, 188 - ac3_unit[ac3_count].Pay_len, ac3_unit[ac3_count].Pay_len);
                                    }
                                    log("Ac3 Adjustment: " + ac3_adjusted.ToString("F3"));
                                }
                            }
                            if (data[i].pid == 0x1100)
                            {
                                if (data[i].pes_st)
                                {
                                    stamps.Write(PTS.ToString("F4") + ">>");
                                    if (new_ac3 != null) 
                                        new_ac3.Write(data[i].pack.data, 188 - data[i].Pay_len + 17, data[i].Pay_len - 17);
                                }
                                else
                                    if (new_ac3 != null) 
                                        new_ac3.Write(data[i].pack.data, 188 - data[i].Pay_len, data[i].Pay_len);
                            }
                            if (data[i].pid == 0x1011)
                            {
                                if (data[i].pes_st)
                                {
                                    stamps.Write(PTS.ToString("F4") + "\r\n");
                                    int afield = bt.ret_bit_value(data[i].pack.data[11], 0, 1);
                                    p = 23;
                                    if (afield == 2)
                                        p = 18;
                                    if (new_264 != null)
                                        new_264.Write(data[i].pack.data, 188 - data[i].Pay_len + p, data[i].Pay_len - p);
                                }
                                else
                                    if (new_264 != null) 
                                        new_264.Write(data[i].pack.data, 188 - data[i].Pay_len, data[i].Pay_len);
                            }

                            pack_nr++;
                        }
                    }
                    if (x < MyBuffer) stop = true;
                    else x = fill_buffer();
                }
                pack_nr -= 2;
                display_pack();
                pack_nr++;
                log("File tsinfo.txt created in program directory");
                log(pack_nr.ToString() + "  packets written ok");
                // not sure what the following intends to do (ts)
                //FileStream f2 = new FileStream(MyTSInfo, FileMode.Open);
                f.Flush();
                log("File size is  " + f.BaseStream.Length.ToString() + " bytes");
                //f2.Close();
            }
            catch (Exception e)
            {
                log("FAILED!:" + e.ToString());
            }
            finally
            {
                if (new_ac3 != null) new_ac3.Close();
                if (new_264 != null) new_264.Close();
                f.Close();
                stamps.Close();
            }
        }

    }
}
